import numpy as np
import pylab as pl
import math
import scipy as sp
from scipy import linalg,random

def gradfix_eps(A,b,u0,rho,eps):
    itmax=50 #20
    n=len(b)
    tabu=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    u0.shape=(n,)
    tabu[:,0]=u0
    u=u0.copy()
    r=b-np.dot(A,u)
    test=sp.linalg.norm(r)
    taberr[0]=test
    it=0
    while test > eps and it < itmax:
        it += 1
        u=u+rho*r
        r=b-np.dot(A,u)
        test=sp.linalg.norm(r)
        u.shape=(n,)
        tabu[:,it]=u
        taberr[it]=test
    return tabu,taberr,it
    
    

def gradopt_eps(A,b,u0,eps):
    n=len(u0)
    itmax=10
    tabu=np.zeros(shape=(n,itmax+1))
    u0.shape=(n,)
    tabu[:,0]=u0
    u=u0.copy()
    r=b-np.dot(A,u)
    test=sp.linalg.norm(r)
    taberr=np.zeros(itmax+1)
    taberr[0]=test
    it=0
    while test > eps and it < itmax:
        it += 1
        lamb=-np.dot(r.T,r)/np.dot(r.T,np.dot(A,r))
        u=u-lamb*r
        r=b-np.dot(A,u)
        test=sp.linalg.norm(r)
        u.shape=(n,)
        tabu[:,it]=u
        taberr[it]=test
    return tabu,taberr,it


def gradfi(fp,x0,rho,eps):
    itmax=100
    tabx=np.zeros(itmax+1)
    taberr=np.zeros(itmax+1)
    pas=fp(x0)
    err=abs(pas)
    x1=x0 #x0.copy()
    tabx[0]=x1
    taberr[0]=err
    it=0
    while err > eps and it < itmax:
        it += 1
        x1 = x1-rho*pas
        pas=fp(x1)
        err=abs(pas)
        tabx[it]=x1
        taberr[it]=err
    return tabx,taberr,it

def gradfiopt(f,fp,x0,eps):
    itmax=100
    rho=np.linspace(-10,10,200)
    tabx=np.zeros(itmax+1)
    taberr=np.zeros(itmax+1)
    pas=fp(x0)
    err=abs(pas)
    x1=x0 #x0.copy()
    tabx[0]=x1
    taberr[0]=err
    it=0
    while err > eps and it < itmax:
        it += 1
        f1=np.array([f(x1-rho[i]*pas) for i in range(200)])
        i0=np.argmin(f1)
        x1 = x1-rho[i0]*pas
        pas=fp(x1)
        err=abs(pas)
        tabx[it]=x1
        taberr[it]=err
    return tabx,taberr,it


def gradfioptb(f,fp,x0,eps):
    itmax=100
    n=len(x0)
    rho=np.linspace(-10,10,200)
    tabx=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    pas=fp(x0)
    err=sp.linalg.norm(pas)
    x1=x0.copy()
    x1.shape=(n,)
    tabx[:,0]=x1
    taberr[0]=err
    it=0
    while err > eps and it < itmax:
        it += 1
        f1=np.array([f(x1-rho[i]*pas) for i in range(200)])
        i0=np.argmin(f1)
        x1 = x1-rho[i0]*pas
        pas=fp(x1)
        err=sp.linalg.norm(pas)
        x1.shape=(n,)
        tabx[:,it]=x1
        taberr[it]=err
    return tabx,taberr,it



def gradfiopt_a(f,jf,fp,jfp,x0,eps,a):
    itmax=100
    n=len(x0)
    rho=np.linspace(-10,10,200)
    tabx=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    pas=fp(x0)+jfp(x0)/a
    err=sp.linalg.norm(pas)
    x1=x0.copy()
    x1.shape=(n,)
    tabx[:,0]=x1
    taberr[0]=err
    it=0
    while err > eps and it < itmax:
        it += 1
        f1=np.array([f(x1-rho[i]*pas)+jf(x1-rho[i]*pas)/a for i in range(200)])
        i0=np.argmin(f1)
        x1 = x1-rho[i0]*pas
        pas=fp(x1)+jfp(x1)/a
        err=sp.linalg.norm(pas)
        x1.shape=(n,)
        tabx[:,it]=x1
        taberr[it]=err
    return tabx,taberr,it

def projK(x):
    y=x.copy()
    x1=x[0]
    x2=x[1]
    s=(x1-x2)/np.sqrt(2)
    t=(x1+x2)/np.sqrt(2)
    r=s**2+t**2
    if r <= 1 and t >= 1/np.sqrt(2):
        s1=s
        t1=t
    if t >= abs(t) and r >= 1:
        s1=s/np.sqrt(r)
        t1=t/np.sqrt(r)
    if t <= 1/np.sqrt(2) and abs(s) <= 1/np.sqrt(2):
        s1=s
        t1=1/np.sqrt(2)
    if t <= abs(s):
        if s >= 1/np.sqrt(2):
            s1=1/np.sqrt(2)
            t1=s1
        if s <= -1/np.sqrt(2):
            s1=-1/np.sqrt(2)
            t1=s1
    y[0]=(s1+t1)/np.sqrt(2)
    y[1]=(t1-s1)/np.sqrt(2)
    return y



def gradfiproj(df,x0,rho,eps):
    itmax=100
    n=len(x0)
    tabx=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    taber=np.zeros(itmax+1)
    it=0
    x=x0.copy()
    x1=projK(x-rho*df(x))
    err=sp.linalg.norm(x1-x)
    taberr[0]=err
    x.shape=(n,)
    tabx[:,0]=x
    while err > eps and it < itmax:
        it += 1
        x=x1.copy()
        x1=projK(x-rho*df(x))
        err=sp.linalg.norm(x1-x)
        taberr[it]=err
        x.shape=(n,)
        tabx[:,it]=x
    for i in range(it):
        taber[i]=sp.linalg.norm(tabx[:,i]-x)
    return tabx,taberr,taber,it

def gradfipen(df,dj,x0,rho,eps,a):
    itmax=100
    n=len(x0)
    tabx=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    taber=np.zeros(itmax+1)
    it=0
    x=x0.copy()
    x1=x-rho*(df(x)+dj(x)/a)
    err=sp.linalg.norm(x1-x)
    taberr[0]=err
    x.shape=(n,)
    tabx[:,0]=x
    while err > eps and it < itmax:
        it += 1
        x=x1.copy()
        x1=x-rho*(df(x)+dj(x)/a)
        err=sp.linalg.norm(x1-x)
        taberr[it]=err
        x.shape=(n,)
        tabx[:,it]=x
    for i in range(it):
        taber[i]=sp.linalg.norm(tabx[:,i]-x)
    return tabx,taberr,taber,it

def gradfi2(fp,x0,rho,eps):
    itmax=500 #200 #100
    n=len(x0)
    tabx=np.zeros(shape=(n,itmax+1))
    taberr=np.zeros(itmax+1)
    pas=fp(x0)
    err=sp.linalg.norm(pas)
    x1=x0.copy()
    x1.shape=(n,)
    tabx[:,0]=x1
    taberr[0]=err
    it=0
    while err > eps and it < itmax:
        it += 1
        x1 = x1-rho*pas
        pas=fp(x1)
        err=sp.linalg.norm(pas)
        tabx[:,it]=x1
        taberr[it]=err
    return tabx,taberr,it

